A simple state machine

----------------------------------------------------------------------

Shallow embedding, interpreted as the current state

> data State = Opened | Closed | Locked deriving (Eq, Enum)
> data Event = Open | Close | Lock | Unlock deriving (Eq, Enum)

> type Door = State

> initial   :: Door
> react     :: Door -> Event -> Door
> isLocked  :: Door -> Bool

> initial = Opened

> react s Open    = if s==Closed  then Opened  else s
> react s Close   = if s==Opened  then Closed  else s
> react s Lock    = if s==Closed  then Locked  else s
> react s Unlock  = if s==Locked  then Closed  else s

> isLocked Locked = True
> isLocked _ = False

----------------------------------------------------------------------

Shallow embedding, as the final coalgebra (infinite behavioural tree)

> data Tree = Node { root :: State, kids :: Event -> Tree }

> type Door' = Tree

> initial'   :: Door'
> react'     :: Door' -> Event -> Door'
> isLocked'  :: Door' -> Bool

> initial' = fetch trees Opened where
>   trees = map construct [Opened .. Locked]
>   construct s = Node s (fetch trees . react s)
>   fetch trees s = trees !! fromEnum s

> react' = kids

> isLocked' = isLocked . root
