A slightly more complicated version of Hutton's Razor:
a heterogeneous expression language.

> {-# LANGUAGE KindSignatures, GADTs #-}

----------------------------------------------------------------------

First, with "dynamic typing".

> data Expr1  =  ValI Integer
>             |  Add Expr1 Expr1
>             |  ValB Bool
>             |  And Expr1 Expr1
>             |  EqZero Expr1
>             |  If Expr1 Expr1 Expr1

> eval :: Expr1 -> Either Integer Bool
> eval (ValI n)    = Left n
> eval (Add x y)   = case (eval x, eval y) of (Left m, Left n) -> Left (m+n)
> eval (ValB b)    = Right b
> eval (And x y)   = case (eval x, eval y) of (Right a, Right b) -> Right (a && b)
> eval (EqZero x)  = case eval x of Left n -> Right (n==0)
> eval (If x y z)  = case eval x of Right b -> if b then eval y else eval z

----------------------------------------------------------------------

An equivalent definition of the datatype, using the GADT style but not
GADT features.

> data Expr2 :: * where
>   ValI2    :: Integer                  -> Expr2
>   Add2     :: Expr2 -> Expr2           -> Expr2
>   ValB2    :: Bool                     -> Expr2
>   And2     :: Expr2 -> Expr2           -> Expr2
>   EqZero2  :: Expr2                    -> Expr2
>   If2      :: Expr2 -> Expr2 -> Expr2  -> Expr2

----------------------------------------------------------------------

Now as a true GADT.

> data Expr3 :: * -> * where
>   ValI3    :: Integer                           -> Expr3 Integer
>   Add3     :: Expr3 Integer -> Expr3 Integer    -> Expr3 Integer
>   ValB3    :: Bool                              -> Expr3 Bool
>   And3     :: Expr3 Bool -> Expr3 Bool          -> Expr3 Bool
>   EqZero3  :: Expr3 Integer                     -> Expr3 Bool
>   If3      :: Expr3 Bool -> Expr3 a -> Expr3 a  -> Expr3 a

> eval3 :: Expr3 a -> a
> eval3 (ValI3 n)    = n
> eval3 (Add3 x y)   = eval3 x + eval3 y
> eval3 (ValB3 b)    = b
> eval3 (And3 x y)   = eval3 x && eval3 y
> eval3 (EqZero3 x)  = eval3 x == 0
> eval3 (If3 x y z)  = if eval3 x then eval3 y else eval3 z

